# Copyright 1999-2016. Parallels IP Holdings GmbH. All Rights Reserved.
""" py-limit-out - Python implementation of Plesk outgoing rate limiting mail handler ('out' part) """

import sys
import logging

from plesk_mail_handlers import LimitHandler, getLogger, LimitHandlerMessages


log = getLogger(__name__, syslog=(__name__ != '__main__'))


class OutLimitHandler(LimitHandler):
    HEADER_SENDER = 'X-PPP-Vhost'

    def preamble(self):
        if "debug" in self.context:
            log.setLevel(logging.DEBUG)

        log.debug("Running py-limit-out handler (context='%s')", self.context)
        self._generate_plesk_message_id = True

    def handle_message(self, db, message, message_id):
        self._log_debug_info(db)

        limits = self._get_limits(db)
        if not limits:
            log.info("No SMTP AUTH and not running in sendmail context (incoming or unrestricted outgoing mail). "
                     "SKIP message.")
            self.exit_skip()

        if self._should_reject_sendmail_usage(limits):
            log.info("REJECT message as system user is not allowed to use sendmail")
            self.reply("554", "5.7.0", LimitHandlerMessages.REJECT_SENDMAIL_NOT_ALLOWED)
            self.exit_reject()

        should_block = self._check_limits_and_update_usage(db, message_id, limits, 'out_limit', 'passed', 'rejected')
        if should_block and "limit" in self.context:
            log.info("REJECT message as outgoing mail rate limits are exceeded")
            self.reply("554", "5.7.0", LimitHandlerMessages.REJECT_OUTGOING_LIMIT_EXCEEDED)
            self.exit_reject()
        else:
            self._enforce_sender(db, limits)
            self._adjust_headers(limits)
            self.exit_pass()

    def _enforce_sender(self, db, limits):
        """ Verify or fix envelope sender and/or sender mail headers.
            This is activated via following outgoing features: {fix,verify}-sender-{envelope,headers}.
        """
        settings = {}
        for feature in self.context.split(','):
            feature_parts = feature.split('-')
            if len(feature_parts) != 3:
                continue
            action, obj, subject = feature_parts
            if obj != "sender" or subject not in ("envelope", "headers") or action not in ("fix", "verify"):
                continue
            settings[subject] = action

        if any(settings.values()):
            self._check_and_update_sender(db, limits, **settings)


    def _adjust_headers(self, limits):
        """ Fix message headers - add external sender vhost header. """
        if "dont-disclose-sender" in self.context:
            return

        del self.message[self.HEADER_SENDER]

        sender = self._get_sender_str(limits)
        log.debug("Setting '%s' header to '%s'", self.HEADER_SENDER, sender)
        if sender:
            self.message[self.HEADER_SENDER] = sender


def main():
    """ py-limit-out mail handler main entry point. """
    try:
        OutLimitHandler().run()
    except SystemExit:
        raise
    except Exception as ex:
        log.error('%s', ex)
        log.debug('This exception happened at:', exc_info=sys.exc_info())
        sys.exit(OutLimitHandler.HOOK_ERROR)

if __name__ == '__main__':
    main()

# vim: ts=4 sts=4 sw=4 et :
